#include <GLES/egl.h>
#include <stdio.h>

/*
 * When a keypress occurs, the thread that is processing messages
 * will set this shared variable indicating which key was most
 * recently pressed.  This variable should use a semaphore for
 * exclusive access but since the message processing thread never
 * reads from this variable this is not critical.  In rare cases,
 * some keypress events will not be recognized by the main
 * rendering thread.
 */
extern unsigned char keypress;

static LRESULT CALLBACK MyWindowProc(HWND hwnd, UINT msg,
                                     WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
    case WM_PAINT:
        {
            PAINTSTRUCT ps;
            HDC hdc = BeginPaint(hwnd, &ps);
            EndPaint(hwnd, &ps);
        }
        return 0;

    case WM_KEYDOWN:
        switch (wParam)
        {
            case VK_ESCAPE:
                PostQuitMessage(0);
                return 0;

            default:
                keypress = (unsigned char) wParam;
                return 0;
        }

    case WM_DESTROY:
        PostQuitMessage(0);
        return 0;
    }

    return DefWindowProc(hwnd, msg, wParam, lParam);
}

static NativeWindowType MyWindow = NULL;
static HANDLE MyWindowThread = NULL;

typedef struct {
    NativeDisplayType disp;
    int x, y, w, h;
    BOOL ready;
} WindowParams;

static DWORD WINAPI MyWindowThreadProc(LPVOID lpParameter)
{
    MSG msg;
    HINSTANCE hinst = (HINSTANCE) GetModuleHandle(NULL);
    WindowParams *params = (WindowParams*) lpParameter;
    HWND hwnd;
    DWORD ws, wsx;
    int dx, dy;

    ws = WS_OVERLAPPED | WS_CAPTION | WS_BORDER | WS_MINIMIZEBOX |
         WS_SYSMENU | WS_VISIBLE;
    wsx = WS_EX_APPWINDOW;
    dx = GetSystemMetrics(SM_CXFIXEDFRAME) * 2;
    dy = GetSystemMetrics(SM_CYFIXEDFRAME) * 2 +
         GetSystemMetrics(SM_CYCAPTION);

    // Register window class.
    {
        WNDCLASS wndclass = {
            CS_OWNDC,                                   // UINT style;
            MyWindowProc,                               // WNDPROC lpfnWndProc;
            0,                                          // int cbClsExtra;
            0,                                          // int cbWndExtra;
            hinst,                                      // HINSTANCE hInstance;
            NULL,                                       // HICON hIcon;
            LoadCursor(NULL, MAKEINTRESOURCE(IDC_ARROW)),// HCURSOR hCursor;
            NULL,                                       // HBRUSH hbrBackground;
            NULL,                                       // LPCTSTR lpszMenuName;
            "MY_WINDOW"                                 // LPCTSTR lpszClassName;
        };

        RegisterClass(&wndclass);
    }

    hwnd = CreateWindowEx(wsx, "MY_WINDOW", "EGL Window", ws,
                          params->x, params->y, params->w + dx, params->h + dy,
                          NULL, NULL, hinst, NULL);
    if (!hwnd)
        return 0;

    // Set global data and give all-clear.
    *((volatile HWND*) &MyWindow) = hwnd;
    *((volatile BOOL*) &params->ready) = TRUE;

    // Run message pump.

    for ( ; ; ) {
        if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
            if (msg.message == WM_QUIT)
                break;
            DispatchMessage(&msg);
            continue;
        }
        WaitMessage();
    }

    // Window has been destroyed.
    MyWindow = NULL;

    //HACK: tear down the window.
    ExitProcess(0);

    return 0;
}

NativeWindowType MyWin32MakeWindow(NativeDisplayType disp,
                                   int x, int y, int w, int h)
{
    WindowParams params;
    DWORD tid = 0;

    // No more than one window at a time. Lame.
    if (MyWindow || MyWindowThread)
        return NULL;

    // Spawn a new thread to create window and run message pump.
    params.disp = disp;
    params.x = x; params.y = y; params.w = w; params.h = h;
    params.ready = FALSE;
    if (!(MyWindowThread = CreateThread(NULL, 0, MyWindowThreadProc,
                                        &params, 0, &tid)))
        return NULL;

    // Wait for new window to be ready, or thread to die.
    do {
        if (WaitForSingleObject(MyWindowThread, 50) == WAIT_OBJECT_0) {
            // Thread terminated; something went wrong.
            CloseHandle(MyWindowThread);
            MyWindow = NULL;
            MyWindowThread = NULL;
            break;
        }
    } while (!*((volatile BOOL *) &params.ready));

    // We're all set.
    return MyWindow;
}
