/*
 * Simple app that demonstrates using per-pixel fog.  The app allows the user
 * to toggle through different fog modes (Off, Linear, Exponential, Exponential^2)
 * to see the impact this has on rendering.  The app also demonstrate how to
 * enable basic lighting using the OpenGL ES api.  This demo makes moderate use of
 * the nvfxmath math library.  The source code demonstrates how to use some of the
 * functions contained in the library.  This code also shows how to toggle between
 * landscape and portrait modes.
 */

#include <GLES/egl.h>
#include <GLES/gl.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include "chalice_fixed.h"
#include "cursor16x16.h"
#include "quat.h"
#include "inverse.h"
#include "nvfxmath/nvfxmath.h"
#ifndef _WIN32
#include "genericinput/genericinput.h"
#endif

static EGLDisplay       dpy;  /* handle to the display          */
static NativeWindowType win;  /* handle to the window           */
static EGLSurface       draw; /* handle to the drawable surface */
static EGLConfig        cfg;  /* handle to an available config  */
static EGLContext       ctx;  /* rendering context              */

static GLuint cursor_tid;     /* the texture handle for the cursor */
static struct {
    int window_height;
    int window_width; 
	int default_rotated;
    int landscape;
    int cursor_flag;
    int cursor_x;
    int cursor_y;
    NvSFx screen_x;
    NvSFx screen_y;
    NvSFx eye_distance;
    NvSFx heading;
    NvSFx pitch;
    NvSFx hv;
    NvSFx pv;
    NvSFx orient[4];
    NvSFx aspect;
    NvSFx near_height;
    NvSFx near_distance;
    NvSFx far_distance;
    int lighting;
    int fog_mode;
    GLfloat fog_color[4];
    int finished;
} state;

/* Amount to move cursor. */
#define CURSOR_DELTA 10


#ifdef _WIN32
extern NativeWindowType MyWin32MakeWindow(NativeDisplayType disp,
                                          int x, int y, int w, int h);
#endif

/*
 * CreateGLWindow - open something to draw on
 */
int CreateGLWindow( int width, int height )
{
    EGLint num_cfgs = 0;
    int attrib[] =
    {
        EGL_RED_SIZE,   1, /* At least 1-bit red.   */
        EGL_GREEN_SIZE, 1, /* At least 1-bit green. */
        EGL_BLUE_SIZE,  1, /* At least 1-bit blue.  */
        EGL_DEPTH_SIZE, 1, /* At least 1-bit depth. */
        EGL_NONE
    };

    /* Get a handle to the default display. */
    dpy = eglGetDisplay( EGL_DEFAULT_DISPLAY );
    if (!dpy)
    {
        fprintf( stderr, "Error: Couldn\'t open the EGL default display\n" );
        return( EGL_FALSE );
    }

    /* Initialize the display before calling all other EGL functions. */
    if ( !eglInitialize( dpy, 0, 0 ) )
    {
        fprintf( stderr, "Error: Couldn\'t init the EGL default display\n" );
        return( EGL_FALSE );
    }

    /* Request an EGL config (i.e. pixel format) that has the requested attributes. */
    if ( !eglChooseConfig( dpy, attrib, &cfg, 1, &num_cfgs ) || (num_cfgs < 1) )
    {
        fprintf( stderr, "Error: Couldn\'t find a config on the EGL default display\n" );
        return( EGL_FALSE );
    }

#ifdef _WIN32
    /* To handle input on Windows, you must create your own window */
    /* with message processing and hand this window off to EGL.    */
    win = MyWin32MakeWindow( EGL_DEFAULT_DISPLAY, 351, 150, width, height );
#else
    /* To create a full screen window, query display width and height */
    /* and use them as parameters when making the window.             */
    width = eglDisplayWidthNV( dpy, cfg );
    height = eglDisplayHeightNV( dpy, cfg );
    win = eglMakeWindowNV( EGL_DEFAULT_DISPLAY, 0, 0, width, height );
#endif
    if (!win)
	{
        fprintf( stderr, "Error: Couldn\'t create window\n" );
        return( EGL_FALSE );
    }

    /* Create the drawable surface and rendering context. */
    draw = eglCreateWindowSurface( dpy, cfg, win, 0 );
    ctx = eglCreateContext( dpy, cfg, 0, 0 );

    /* Make the context current and bind the newly created drawable */
    /* surface as the current "draw" and "read" surface.            */
    eglMakeCurrent( dpy, draw, draw, ctx );

    return( EGL_TRUE );
}

/*
 * KillGLWindow - shut down rendering
 */
void KillGLWindow()
{
    if (dpy)
    {
        /* Delete created resources. */
        if (ctx) eglDestroyContext( dpy, ctx );
        if (draw) eglDestroySurface( dpy, draw );
        if (win) eglDestroyWindowNV( win );

        /* Release resources associated with the display. */
        eglTerminate( dpy );
    }
}

static void CleanUp()
{
	KillGLWindow();
}

#ifndef _WIN32
static void FatalError(const char *msg, ...)
     __attribute__ ((format (printf, 1, 2), noreturn));
#else
DECLSPEC_NORETURN static void FatalError(const char *msg, ...);
#endif

static void FatalError(const char *msg, ...)
{
    va_list ap;
#ifdef _WIN32
    char error[512];
#endif
    va_start(ap, msg);
    vfprintf(stderr, msg, ap);
    va_end(ap);
    fputc('\n', stderr);
#ifdef _WIN32
    va_start(ap, msg);
    _vsnprintf( error, sizeof(error), msg, ap ); 
    va_end(ap);
    MessageBox( NULL, error, "FatalError Message", MB_OK );
#endif
    CleanUp();
    exit(-1);
}

static void CheckError(const char *msg)
{
    int e = glGetError();
    if (e) FatalError("OpenGL ES Error %s: 0x%.8x", msg, e);
}


static void CursorMotion( int x, int y )
{
    NvSFx window_x, window_y;
    NvSFx halfh, halfw;

    if ((x > 0) && (x < state.window_width))
        state.cursor_x = x;
    if ((y > 0) && (y < state.window_height))
        state.cursor_y = y;
    
    halfw = (state.window_width >> 1);  /* holds integer width/2  */
    halfh = (state.window_height >> 1); /* holds integer height/2 */
    if ( !state.landscape )
	{
        window_y = (state.window_height - state.cursor_y) - halfh;
        window_y = (window_y << 16); /* convert to NvSFx */
        halfh    = (halfh << 16);    /* convert to NvSFx */
        state.screen_y = NvSFxDiv( window_y, halfh );
        window_x = state.cursor_x - halfw;
        window_x = (window_x << 16); /* convert to NvSFx */
        halfw    = (halfw << 16);    /* convert to NvSFx */
        state.screen_x = NvSFxDiv( window_x, halfw );
	}
	else
	{
        window_y = (state.window_height - state.cursor_y) - halfh;
        window_y = (window_y << 16); /* convert to NvSFx */
        halfh    = (halfh << 16);    /* convert to NvSFx */
        state.screen_x = NvSFxDiv( -window_y, halfh );
        window_x = state.cursor_x - halfw;
        window_x = (window_x << 16); /* convert to NvSFx */
        halfw    = (halfw << 16);    /* convert to NvSFx */
        state.screen_y = NvSFxDiv( window_x, halfw );
	}
}


/*
 * On Windows, when a keypress occurs, the thread that is processing
 * messages will set this shared variable indicating which key was
 * most recently pressed.  This variable should use a semaphore for
 * exclusive access but since the message processing thread never
 * reads from this variable this is not critical.  In rare cases,
 * some keypress events will not be recognized by the main
 * rendering thread.
 *
 * There are alternative ways to handle keyboard input using this
 * EGL on windows implementation.  Please see the win32ui.c file in
 * this project to see the details of what the message processing
 * thread does.
 */
unsigned char keypress = 255;

void keyboard(unsigned char key)
{
    switch( key )
    {
        case ' ':
            break;
        case 36:
        case 103: /* '7' <Home> key */
            state.lighting = !state.lighting;
            break;
        case 33:
        case 105: /* '9' <Pg Up> key */
            state.fog_mode = (state.fog_mode+1) %4;
            switch ( state.fog_mode )
            {
                case 0:
                    /* Fog is disabled. */
                    break;
                case 1:
                    /* Fog is linear. */
                    glFogf( GL_FOG_MODE, GL_LINEAR );
                    break;
                case 2:
                    /* Fog factor f = e^(-density*z) */
                    glFogf( GL_FOG_MODE, GL_EXP );
                    break;
                case 3:
                    /* Fog factor f = e^((-density*z)^2) */
                    glFogf( GL_FOG_MODE, GL_EXP2 );
                    break;
            }
            break;
        case 27:
            state.finished = 1;
            break;
        case 45:
        case 96: /* '0' <Ins> key */
            state.landscape = !state.landscape;
            break;
        case '8':
        case 38:
        case 104: /* Up Arrow */
			/* Assumes upper left corner is (0,0) */
			if ( state.landscape )
			    CursorMotion(state.cursor_x+CURSOR_DELTA, state.cursor_y);
			else
			    CursorMotion(state.cursor_x, state.cursor_y-CURSOR_DELTA);
			return;

        case '2':
        case 40:
        case 98: /* Down Arrow */
			/* Assumes upper left corner is (0,0) */
			if ( state.landscape )
			    CursorMotion(state.cursor_x-CURSOR_DELTA, state.cursor_y);
			else
			    CursorMotion(state.cursor_x, state.cursor_y+CURSOR_DELTA);
			break;

        case '4':
        case 37:
        case 100: /* Left Arrow */
			if ( state.landscape )
			    CursorMotion(state.cursor_x, state.cursor_y-CURSOR_DELTA);
			else
			    CursorMotion(state.cursor_x-CURSOR_DELTA, state.cursor_y);
			return;

        case '6':
        case 39:
        case 102: /* Right Arrow */
			if ( state.landscape )
			    CursorMotion(state.cursor_x, state.cursor_y+CURSOR_DELTA);
			else
			    CursorMotion(state.cursor_x+CURSOR_DELTA, state.cursor_y);
			return;
    }
}

#ifndef _WIN32
static void GetKeyboardInput()
{
	unsigned char c;
	if ( readRawKeyboard(&c) > 0)
	{
	    keyboard( c );
	}
}
#endif

static void InitGraphicsState()
{
    /* We will define a single-light at (0,0,0) in eye space. */
    GLfixed light_position[4] = { 0, 0, 0, 0x00010000 };

    /* Initialize all the state variables. */
    eglQuerySurface( dpy, draw, EGL_WIDTH, &state.window_width );
    eglQuerySurface( dpy, draw, EGL_HEIGHT, &state.window_height );
    state.cursor_flag = 1;
    state.cursor_x = state.window_width >> 1;
    state.cursor_y = state.window_height >> 1;
    state.screen_x = 0x0;
    state.screen_y = 0x0;

    if ( state.window_width > state.window_height )
    {
        state.default_rotated = 1;
        state.landscape = 1;
    }
    else
    {
        state.default_rotated = 0;
        state.landscape = 0;
    }

    /* Initialize the frustum parameters. */
    state.eye_distance = 0x00080000;  /* eye distance to focus point = 8.0  */
    state.near_height = 0x00010000;   /* height on the near plane    = 1.0  */ 
    state.near_distance = 0x00020000; /* distance to the near plane  = 2.0  */
    state.far_distance = 0x00280000;  /* distance to the far plane   = 40.0 */

    /* Initialize these to 0. */
    state.heading = 0x00000000;
    state.pitch = 0x00000000;
    state.hv = 0x00000000;
    state.pv = 0x00000000;

    /* Initially want lighting turned on. */
    state.lighting = 1;

    /* Set the initial fog mode and color. */
    state.fog_color[0] = 0.2f;
    state.fog_color[1] = 0.2f;
    state.fog_color[2] = 1.0f;
    state.fog_color[3] = 1.0f;
    state.fog_mode = 2;

    /* This flag will take on a value of 1 when time to shutdown the app. */
    state.finished = 0;

    /* Initialize the cursor texture. */
    glGenTextures( 1, &cursor_tid );

    glBindTexture( GL_TEXTURE_2D, cursor_tid );
    glTexImage2D( GL_TEXTURE_2D, 0, GL_ALPHA,
                            cursor_res, cursor_res, 0, GL_ALPHA, GL_UNSIGNED_BYTE,
                            cursor_texdata );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

    /* The chalice model has front faces that are clockwise winding. */
    glFrontFace( GL_CW );
    glCullFace( GL_BACK );

    /* Enable depth test. */
    glDepthFunc( GL_LEQUAL );
    glEnable( GL_DEPTH_TEST );

    /* Setup a single light source and enable it. */
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    glLightxv( GL_LIGHT0, GL_POSITION, light_position );
    glEnable( GL_LIGHT0 );

    /* Setup the fog. */
    glClearColor( 0.2f, 0.2f, 1.0f, 1.0f );
    glFogfv( GL_FOG_COLOR, state.fog_color );
    /*  glFogf( GL_FOG_DENSITY, 0.35f ); */
    glFogf( GL_FOG_DENSITY, 0.25f );
    glHint( GL_FOG_HINT, GL_NICEST );
    glFogf( GL_FOG_START, 4.0 );
    glFogf( GL_FOG_END, 9.0 );
    glFogf( GL_FOG_MODE, GL_EXP );

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	CheckError( "InitGraphicsState" );
}

static void DrawCursor(void)
{
    static float texcoords[2*4] = 
    { 
       0.0f, 0.0f,
       0.0f, 1.0f,
       1.0f, 0.0f,
       1.0f, 1.0f 
    };
	GLfloat square[2*4];
	GLfloat halfw;
	GLfloat halfh;
    GLfloat x, y;

    glColor4f( 1.0, 1.0, 1.0, 0.0 );
    glDisable( GL_DEPTH_TEST );

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity(); 
	halfw = (GLfloat)(state.window_width >> 1);
	halfh = (GLfloat)(state.window_height >> 1);
    glOrthof( -halfw, halfw, -halfh, halfh, 0, 1 );

	x = state.cursor_x - halfw;
	y = (state.window_height - state.cursor_y) - halfh;  /* Assumes upper left corner is (0,0) */

    square[2] = x - 8;
    square[3] = y - 8;
    square[0] = x - 8;
    square[1] = y + 8;
    square[4] = x + 8;
    square[5] = y + 8;
    square[6] = x + 8;
    square[7] = y - 8;

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity(); 

    glEnableClientState ( GL_VERTEX_ARRAY );
	glVertexPointer( 2, GL_FLOAT, 0, square );
	glEnableClientState ( GL_TEXTURE_COORD_ARRAY );
	glTexCoordPointer( 2, GL_FLOAT, 0, texcoords );

    glEnable( GL_TEXTURE_2D );
    glBindTexture( GL_TEXTURE_2D, cursor_tid );

    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
    glEnable( GL_ALPHA_TEST );
    glAlphaFunc( GL_NOTEQUAL, 0.0f );
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
        glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
    glDisable( GL_BLEND );
    glDisable( GL_ALPHA_TEST );

    glDisable( GL_TEXTURE_2D );
    glDisableClientState( GL_VERTEX_ARRAY );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );

    glEnable( GL_DEPTH_TEST );
}

static void ComputeEyePosition( NvSFx v[3] )
{
#if 0
    NvSFx pitch_axis[3] = { 0x00010000, 0x00000000, 0x00000000 };
    NvSFx heading_axis[3] = { 0x00000000, 0x00010000, 0x00000000 };
#else
    NvSFx pitch_axis[3] = { -0x00010000, 0x00000000, 0x00000000 };
    NvSFx heading_axis[3] = { 0x00000000, 0x00000000, -0x00010000 };
#endif

    NvSFx tempQuat1[4];
    NvSFx tempQuat2[4];
    NvSFx m[4][4];
    NvSFx minv[4][4];

    MakeQuat( tempQuat1, state.pitch, pitch_axis );
    MakeQuat( tempQuat2, state.heading, heading_axis );
    MulQuat( state.orient, tempQuat1, tempQuat2 );
    QuatToMat( m, state.orient );

    inverse( minv, m );

    v[0] = NvSFxMul( -minv[2][0], state.eye_distance );
    v[1] = NvSFxMul( -minv[2][1], state.eye_distance );
    v[2] = NvSFxMul( -minv[2][2], state.eye_distance );
}
 
static void SetAspect()
{
    state.aspect = 0x00015555;  /* set aspect to 1.33333 */

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity(); 

    glFrustumx(-state.near_height,
               state.near_height,
               NvSFxMul( -state.near_height, state.aspect ),
               NvSFxMul(  state.near_height, state.aspect ), state.near_distance, state.far_distance );

    /* Rotate geometry so that largest dimension is horizontal. */
	if ( state.landscape )
        glRotatef(-90, 0, 0, 1);
}

static void Display(void)
{
    NvSFx m[4][4];
    NvSFx EyePos[3];

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    SetAspect();

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();

    /* Apply the viewing transformation. */
    QuatToMat( m, state.orient );
    glMultMatrixx( (GLfixed *) &m[0][0]);
    ComputeEyePosition( EyePos );
    glTranslatex( EyePos[0], EyePos[1], EyePos[2] );

    /* Apply the modelling transformation. */
    glTranslatex( -modelcenter[0], -modelcenter[1], -modelcenter[2] );

    /* Turn on backface culling. */
    glEnable( GL_CULL_FACE );

    /* Turn on fog. */
    if ( state.fog_mode != 0 )
        glEnable( GL_FOG );

    /* Turn on lighting. */
    if ( state.lighting )
        glEnable( GL_LIGHTING );
 
    glEnableClientState( GL_VERTEX_ARRAY );
    glVertexPointer( 3, GL_FIXED, 0, (void *) vertexdata );
    glEnableClientState( GL_NORMAL_ARRAY );
    glNormalPointer( GL_FIXED, 0, (void *) normaldata );

    glDrawElements( GL_TRIANGLES, nindices, GL_UNSIGNED_SHORT, indices );

    glDisableClientState( GL_NORMAL_ARRAY );
    glDisableClientState( GL_VERTEX_ARRAY );

    /* Turn off lighting. */
    glDisable( GL_LIGHTING );

    /* Turn off fog. */
    glDisable( GL_FOG );

    /* Turn off back face culling. */
    glDisable( GL_CULL_FACE );

    /* Draw the cursor. */
    if ( state.cursor_flag )
        DrawCursor();

    eglSwapBuffers( dpy, draw );
}

static void tick(void)
{
    /* Update the rotation. */
    /*
    state.hv += state.screen_x * 0.001f;
    state.pv += state.screen_y * 0.001f;
    state.pv -= state.pitch * 0.0006f;
    state.hv *= fix(0.94f);
    state.pv *= fix(0.94f);

    state.heading += state.hv;
    state.pitch += state.pv;

    if (state.heading>M_PI) state.heading -=  (fix) (2.0f*M_PI);
    if (state.heading<-M_PI) state.heading += (fix) (2.0f*M_PI);
    */

    state.hv = NvSFxAdd( state.hv, NvSFxMul( state.screen_x, 0x00000041 ) );
    state.pv = NvSFxAdd( state.pv, NvSFxMul( state.screen_y, 0x00000041 ) );
    state.pv = NvSFxSub( state.pv, NvSFxMul( state.pitch, 0x00000027 ) );
    state.hv = NvSFxMul( state.hv, 0x0000f0a3 );
    state.pv = NvSFxMul( state.pv, 0x0000f0a3 );
    state.heading = NvSFxAdd( state.heading, state.hv );
    state.pitch = NvSFxAdd( state.pitch, state.pv );

    if ( state.heading > 0x0003243F ) state.heading = NvSFxSub( state.heading, 0x0006487E );
    if ( state.heading < 0xfffCDBC1 ) state.heading = NvSFxAdd( state.heading, 0x0006487E );

    Display();
}

int main()
{
	if ( !CreateGLWindow( 240, 320 ) )
	{
		FatalError( "Display initialization failed.\nExiting." );
	}

    InitGraphicsState();

#ifndef _WIN32
    initRawKeyboard();
#endif

    /* Print Usage */
    fprintf( stderr, "\n" );
    fprintf( stderr, "Fog Keyboard Controls:\n" );
    fprintf( stderr, "---------------------\n" );
    fprintf( stderr, "    keypad 8 -- Move mouse cursor up\n" );
    fprintf( stderr, "    keypad 2 -- Move mouse cursor down\n" );
    fprintf( stderr, "    keypad 4 -- Move mouse cursor left\n" );
    fprintf( stderr, "    keypad 6 -- Move mouse cursor right\n" );
    fprintf( stderr, "    keypad 9 -- Cycle through fog modes\n" );
    fprintf( stderr, "    keypad 7 -- Toggle lighting on/off\n" );
    fprintf( stderr, "    keypad 0 -- Toggle viewing orientation\n" );
    fprintf( stderr, "    <ESC>    -- Quit\n" );

    while ( !state.finished )
    {
#ifndef _WIN32
        GetKeyboardInput();
#endif
        keyboard( keypress );
        keypress = 255;
        tick();
    }

#ifndef _WIN32
	    closeRawKeyboard();
#endif

    CleanUp();

	return 0;
}

