/*
 * This app demonstrates using multi-texturing and texture combining.
 * The demo allows the user to cycle through different texture blending
 * to better understand how multi-texturing and texture combining work.
 * The code demonstrates how to load/use DXT1 compressed textures,
 * how to set texture filtering/wrap modes, and how to use a texture
 * matrix for modifying texture coordinates.  It also shows how the
 * application programmer can query for OpenGL ES extensions.
 */

#include <GLES/egl.h>
#include <GLES/gl.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include "nvfxmath/nvfxmath.h"
#include "nvdds/nvdds.h"
#include "nvmediapath/nvmediapath.h"
#include "nvtexture/nvtexture.h"
#ifndef _WIN32
#include "genericinput/genericinput.h"
#endif

static EGLDisplay       dpy;  /* handle to the display          */
static NativeWindowType win;  /* handle to the window           */
static EGLSurface       draw; /* handle to the drawable surface */
static EGLConfig        cfg;  /* handle to an available config  */
static EGLContext       ctx;  /* rendering context              */

/* Number of different texture environment combine modes. */
#define NUM_TEXENV_MODES 8

/* Macros for conversion to fixed point. */
#define FIXINT(x) (x << 16)
#define FIXHALF (FIXINT(1) >> 1)

static int combinemode = 5;         /* Initial combine mode. */
static int finished = 0;            /* Flag indicating if the app should exit. */

static int translate_texture0 = 1;  /* Flag indicating if tex0 should translate. */
static int rotate_texture1 = 1;     /* Flag indicating if tex1 should rotate.    */
static NvSFx tx0rot = FIXINT(90);   /* Initial rotation amount. */
static NvSFx tx1rot = FIXINT(0);    /* Initial rotation amount. */

static GLint window_width;          /* Window width. */
static GLint window_height;         /* Window height. */
static GLuint texid[2];             /* Texture ids for the two textures. */
static GLuint stringtex[NUM_TEXENV_MODES];/* Texture ids for the "text" textures. */

/* Geometry. */
static GLubyte ind [4] = { 0, 1, 2, 3 };
static GLfixed vert[8] = { -FIXHALF,  FIXHALF,
                           -FIXHALF, -FIXHALF,
                            FIXHALF,  FIXHALF,
                            FIXHALF, -FIXHALF };
static GLfixed tex [8] = { FIXINT(0), FIXINT(1),
                           FIXINT(0), FIXINT(0),
                           FIXINT(1), FIXINT(1),
                           FIXINT(1), FIXINT(0) };
static GLfixed col [16] = { FIXINT(1), FIXINT(0), FIXINT(0), FIXINT(1),
                            FIXINT(0), FIXINT(1), FIXINT(0), FIXINT(1),
                            FIXINT(0), FIXINT(0), FIXINT(1), FIXINT(1),
                            FIXINT(1), FIXINT(0), FIXINT(1), FIXINT(1) };
static GLfixed textvert[8] = { -FIXINT(1),  FIXINT(1),
                               -FIXINT(1), -FIXINT(1),
                                FIXINT(1),  FIXINT(1),
                                FIXINT(1), -FIXINT(1) };




#ifdef _WIN32
extern NativeWindowType MyWin32MakeWindow(NativeDisplayType disp,
                                          int x, int y, int w, int h);
#endif

/*
 * CreateGLWindow - open something to draw on
 */
int CreateGLWindow( int width, int height )
{
    EGLint num_cfgs = 0;
    int attrib[] =
    {
        EGL_RED_SIZE,   1, /* At least 1-bit red.   */
        EGL_GREEN_SIZE, 1, /* At least 1-bit green. */
        EGL_BLUE_SIZE,  1, /* At least 1-bit blue.  */
        EGL_DEPTH_SIZE, 1, /* At least 1-bit depth. */
        EGL_NONE
    };

    /* Get a handle to the default display. */
    dpy = eglGetDisplay( EGL_DEFAULT_DISPLAY );
    if (!dpy)
    {
        fprintf( stderr, "Error: Couldn\'t open the EGL default display\n" );
        return( EGL_FALSE );
    }

    /* Initialize the display before calling all other EGL functions. */
    if ( !eglInitialize( dpy, 0, 0 ) )
    {
        fprintf( stderr, "Error: Couldn\'t init the EGL default display\n" );
        return( EGL_FALSE );
    }

    /* Request an EGL config (i.e. pixel format) that has the requested attributes. */
    if ( !eglChooseConfig( dpy, attrib, &cfg, 1, &num_cfgs ) || (num_cfgs < 1) )
    {
        fprintf( stderr, "Error: Couldn\'t find a config on the EGL default display\n" );
        return( EGL_FALSE );
    }

#ifdef _WIN32
    /* To handle input on Windows, you must create your own window */
    /* with message processing and hand this window off to EGL.    */
    win = MyWin32MakeWindow( EGL_DEFAULT_DISPLAY, 351, 150, width, height );
#else
    /* To create a full screen window, query display width and height */
    /* and use them as parameters when making the window.             */
    width = eglDisplayWidthNV( dpy, cfg );
    height = eglDisplayHeightNV( dpy, cfg );
    win = eglMakeWindowNV( EGL_DEFAULT_DISPLAY, 0, 0, width, height );
#endif
    if (!win)
	{
        fprintf( stderr, "Error: Couldn\'t create window\n" );
        return( EGL_FALSE );
    }

    /* Create the drawable surface and rendering context. */
    draw = eglCreateWindowSurface( dpy, cfg, win, 0 );
    ctx = eglCreateContext( dpy, cfg, 0, 0 );

    /* Make the context current and bind the newly created drawable */
    /* surface as the current "draw" and "read" surface.            */
    eglMakeCurrent( dpy, draw, draw, ctx );

    return( EGL_TRUE );
}

/*
 * KillGLWindow - shut down rendering
 */
void KillGLWindow()
{
    if (dpy)
    {
        /* Delete created resources. */
        if (ctx) eglDestroyContext( dpy, ctx );
        if (draw) eglDestroySurface( dpy, draw );
        if (win) eglDestroyWindowNV( win );

        /* Release resources associated with the display. */
        eglTerminate( dpy );
    }
}

static void CleanUp()
{
	KillGLWindow();
}

#ifndef _WIN32
static void FatalError(const char *msg, ...)
     __attribute__ ((format (printf, 1, 2), noreturn));
#else
DECLSPEC_NORETURN static void FatalError(const char *msg, ...);
#endif

static void FatalError(const char *msg, ...)
{
    va_list ap;
#ifdef _WIN32
    char error[512];
#endif
    va_start(ap, msg);
    vfprintf(stderr, msg, ap);
    va_end(ap);
    fputc('\n', stderr);
#ifdef _WIN32
    va_start(ap, msg);
    _vsnprintf( error, sizeof(error), msg, ap ); 
    va_end(ap);
    MessageBox( NULL, error, "FatalError Message", MB_OK );
#endif
    CleanUp();
    exit(-1);
}

static void CheckError(const char *msg)
{
    int e = glGetError();
    if (e) FatalError("OpenGL ES Error %s: 0x%.8x", msg, e);
}


/*
 * On Windows, when a keypress occurs, the thread that is processing
 * messages will set this shared variable indicating which key was
 * most recently pressed.  This variable should use a semaphore for
 * exclusive access but since the message processing thread never
 * reads from this variable this is not critical.  In rare cases,
 * some keypress events will not be recognized by the main
 * rendering thread.
 *
 * There are alternative ways to handle keyboard input using this
 * EGL on windows implementation.  Please see the win32ui.c file in
 * this project to see the details of what the message processing
 * thread does.
 */
unsigned char keypress = 255;

void keyboard(unsigned char key)
{
    switch( key )
    {
        case 27:
            finished = 1;
            break;

        case 38:
        case 104: /* Up Arrow */
            combinemode = 7;
            break;

        case 40:
        case 98: /* Down Arrow */
            combinemode = 1;
            break;

        case 37:
        case 100: /* Left Arrow */
            combinemode = 3;
            break;

        case 39:
        case 102: /* Right Arrow */
            combinemode = 5;
            break;

        case 12:
        case 101: /* '5' key */
            combinemode = 4;
            break;

        case 36:
        case 103: /* '7' <Home> key */
            combinemode = 6;
            break;

        case 33:
        case 105: /* '9' <Pg Up> key */
            combinemode = (combinemode+1) % 8;
            break;

        case 35:
        case 97: /* '1' <End> key */
            combinemode = 0;
            break;

        case 34:
        case 99: /* '3' <Pg Dn> key */
            combinemode = 2;
            break;

        case 45:
        case 96: /* '0' <Ins> key */
            translate_texture0 = !translate_texture0;
            break;

        case 46:
        case 110: /* '.' <Del> key */
            rotate_texture1 = !rotate_texture1;
            break;
    }
}

#ifndef _WIN32
static void GetKeyboardInput()
{
	unsigned char c;
	if ( readRawKeyboard(&c) > 0)
	{
	    keyboard( c );
	}
}
#endif

static void InitGraphicsState()
{
    char filename[256];
    char pathname[256];
    char fullpathname[512];
    char *textname[NUM_TEXENV_MODES] = {
               "multitex_none.dds", "multitex_single.dds",
               "multitex_reprep.dds", "multitex_repmod.dds",
               "multitex_repblend.dds", "multitex_repadd.dds",
               "multitex_modadd.dds", "multitex_blendadd.dds" };
    int i;

    /* Retrieve the size of the drawable surface. */
    eglQuerySurface( dpy, draw, EGL_WIDTH, &window_width );
    eglQuerySurface( dpy, draw, EGL_HEIGHT, &window_height );

    /* Load the textures. */
    glGenTextures( 2, texid );

    glBindTexture( GL_TEXTURE_2D, texid[0] );
    sprintf( pathname, "MEDIA/textures/2D/" );
    sprintf( filename, "numbers.dds" );
    LocateMediaFile( filename, pathname, fullpathname );
    if ( !LoadTextureDDS( fullpathname ) )
        FatalError( "InitGraphicsState()" ); 
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );

    glBindTexture( GL_TEXTURE_2D, texid[1] );
    sprintf( pathname, "MEDIA/textures/2D/" );
    sprintf( filename, "nvlogo_spot_falloff.dds" );
    LocateMediaFile( filename, pathname, fullpathname );
    if ( !LoadTextureDDS( fullpathname ) )
        FatalError( "InitGraphicsState()" ); 
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

    /* Load the "text" textures. */
    glGenTextures( NUM_TEXENV_MODES, stringtex );
    for ( i = 0; i < NUM_TEXENV_MODES; i++ )
    {
        glBindTexture( GL_TEXTURE_2D, stringtex[i] );
        sprintf( pathname, "MEDIA/textures/2D/" );
        sprintf( filename, "%s", textname[i] );
        LocateMediaFile( filename, pathname, fullpathname );
        if ( !LoadTextureDDS( fullpathname ) )
            FatalError( "InitGraphicsState()" ); 
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
    }

    glClearColorx(0, 0, 0, 0);
    glOrthox(-FIXINT(1), FIXINT(1), -FIXINT(1),
             FIXINT(1), -FIXINT(1), FIXINT(1));
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	CheckError( "InitGraphicsState" );
}

void DisplayDualTextureQuad( GLfloat param0, GLfloat param1 )
{
    GLfixed tranx;
    GLfixed trany;

	/*
     * Frivolous use of the texture matrix is not advised
	 * if T&L performance is a bottleneck.
     */
	if ( translate_texture0 )
	{
        /* Compute a circular translation */
		tranx = NvSFxSinDFullRange( tx0rot );
		trany = NvSFxCosDFullRange( tx0rot );

        /* Scale the translation amount by 0.125 */
        tranx = NvSFxMul( 0x00002000, tranx );
        trany = NvSFxMul( 0x00002000, trany );

		glActiveTexture( GL_TEXTURE0 );
        glMatrixMode( GL_TEXTURE );
		glLoadIdentity();
		glTranslatex( tranx, trany, 0 );
		glMatrixMode( GL_MODELVIEW );
	}

	if ( rotate_texture1 )
	{
		glActiveTexture( GL_TEXTURE1 );
		glMatrixMode( GL_TEXTURE );
		glLoadIdentity();

        /*
           Apply a rotation transformation.
           Translate by (0.5, 0.5),
           Rotate about (0,0,1),
           Translate by (-0.5,-0.5)
         */
		glTranslatex( 0x00008000, 0x00008000, 0 );
		glRotatex( tx1rot, 0x00000000, 0x00000000, 0x00010000 );
		glTranslatex( 0xFFFF8000, 0xFFFF8000, 0 );
		glMatrixMode( GL_MODELVIEW );
		glActiveTexture( GL_TEXTURE0 );
	}

    glEnableClientState( GL_VERTEX_ARRAY );
    glEnableClientState( GL_COLOR_ARRAY );
    glVertexPointer( 2, GL_FIXED, 0, (void *) vert );
    glColorPointer( 4, GL_FIXED, 0, (void *) col );

    /* param0 with value of -1 means draw w/o texture */
    if ( param0 != -1 )
    {
        glClientActiveTexture( GL_TEXTURE0 );
        glEnableClientState( GL_TEXTURE_COORD_ARRAY );
        glTexCoordPointer( 2, GL_FIXED, 0, tex );

        glActiveTexture( GL_TEXTURE0 );
        glBindTexture( GL_TEXTURE_2D, texid[0] );
        glEnable( GL_TEXTURE_2D );
        glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, param0 );

        /* param1 with value of -1 means only draw single-textured */
        if ( param1 != -1 )
        {
            glClientActiveTexture( GL_TEXTURE1 );
            glEnableClientState( GL_TEXTURE_COORD_ARRAY );
            glTexCoordPointer( 2, GL_FIXED, 0, tex );

            glActiveTexture( GL_TEXTURE1 );
            glBindTexture( GL_TEXTURE_2D, texid[1] );
            glEnable( GL_TEXTURE_2D );
            glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, param1 );
        }
    }

    glDrawElements( GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_BYTE, ind );

    /* only disable what we turned on */
    if ( param0 != -1 )
    {
        if ( param1 != -1 )
        {
            glClientActiveTexture( GL_TEXTURE1 );
            glDisableClientState( GL_TEXTURE_COORD_ARRAY );
            glActiveTexture( GL_TEXTURE1 );
            glDisable( GL_TEXTURE_2D );
        }
        glClientActiveTexture( GL_TEXTURE0 );
        glDisableClientState( GL_TEXTURE_COORD_ARRAY );
        glActiveTexture( GL_TEXTURE0 );
        glDisable( GL_TEXTURE_2D );
    }

    glDisableClientState( GL_COLOR_ARRAY );
    glDisableClientState( GL_VERTEX_ARRAY );

    /* Set the texture matrices to identity. */
	if ( rotate_texture1 )
	{
		glActiveTexture( GL_TEXTURE1 );
		glMatrixMode( GL_TEXTURE );
		glLoadIdentity();
    }
	if ( translate_texture0 )
	{
 		glActiveTexture( GL_TEXTURE0 );
        glMatrixMode( GL_TEXTURE );
		glLoadIdentity();
   }
	glMatrixMode( GL_MODELVIEW );
}

void DisplayText( int index )
{
	glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();

    glEnableClientState( GL_VERTEX_ARRAY );
    glVertexPointer( 2, GL_FIXED, 0, (void *) textvert );

    glActiveTexture( GL_TEXTURE0 );
    glBindTexture( GL_TEXTURE_2D, stringtex[index] );
    glEnable( GL_TEXTURE_2D );
    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );

    glClientActiveTexture( GL_TEXTURE0 );
    glEnableClientState( GL_TEXTURE_COORD_ARRAY );
    glTexCoordPointer( 2, GL_FIXED, 0, tex );

    glDrawElements( GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_BYTE, ind );

    glClientActiveTexture( GL_TEXTURE0 );
    glDisableClientState( GL_TEXTURE_COORD_ARRAY );
    glActiveTexture( GL_TEXTURE0 );
    glDisable( GL_TEXTURE_2D );

    glDisableClientState( GL_VERTEX_ARRAY );
}

static void Display(void)
{
    GLfloat mode0;
    GLfloat mode1;
    GLfloat Ccolor[4] = { 0.5f, 0.5f, 0.5f, 1.0f };

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glViewport( 20, 0, 280, 280 );
    switch ( combinemode )
    {
        case 0:
            /* C = Cf */
            mode0 = -1;
            mode1 = -1;
            break;

        case 1:
            /* C = C0 */
            mode0 = GL_REPLACE;
            mode1 = -1;
            break;

        case 2:
            /* C = C1 */
            mode0 = GL_REPLACE;
            mode1 = GL_REPLACE;
            break;
        case 3:
            /* C = C0C1 */
            mode0 = GL_REPLACE;
            mode1 = GL_MODULATE;
            break;
        case 4:
            /* C = C0(1-C1) + CcC1 */
            mode0 = GL_REPLACE;
            mode1 = GL_BLEND;
            /* Explicitly set the Cc color. */ 
            glTexEnvfv( GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, Ccolor ); 
            break;
        case 5:
            /* C = C0 + C1 */
            mode0 = GL_REPLACE;
            mode1 = GL_ADD;
            break;
        case 6:
            /* C = CfC0 + C1 */
            mode0 = GL_MODULATE;
            mode1 = GL_ADD;
            break;
        case 7:
            /* C = Cf(1-C0) + CcC0 + C1 */
            mode0 = GL_BLEND;
            mode1 = GL_ADD;
            /* Explicitly set the Cc color. */ 
            glTexEnvfv( GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, Ccolor ); 
            break;
    }
	DisplayDualTextureQuad( mode0, mode1 );
    glViewport( 32, 0, 256, 64 );   /* 256x64 is size of text texture */
    DisplayText( combinemode );

	glViewport( 0, 0, window_width, window_height );

    eglSwapBuffers( dpy, draw );
}

static void tick(void)
{
    if ( translate_texture0 )
    {
        /* Add ~1.00001 degrees each frame. */
        tx0rot = NvSFxAdd( tx0rot, 0x00010001 );
    }
    if ( rotate_texture1 )
	{
        /* Add ~1.50001 degrees each frame. */
        tx1rot = NvSFxAdd( tx1rot, 0x00018001 );
    }

    Display();
}

static GLboolean QueryExtension(char *extName)
{
    unsigned int n;
    char *p = (char *) glGetString(GL_EXTENSIONS);
    char *end = p + strlen(p); 
    while ( p < end )
    {
        n = (unsigned int) strcspn(p, " ");
        if ((strlen(extName)==n) && (strncmp(extName,p,n)==0)) {
            return GL_TRUE;
        }
        p += (n + 1);
    }
    return GL_FALSE;
}

void VerifyExtensionsSupported( char **extensionlist, int n )
{
    int i;

    for ( i = 0; i < n; i++ )
    {
        if ( !QueryExtension( extensionlist[i] ) )
            FatalError( "Extension: %s not supported but is required to run this demo.\nExiting.", extensionlist[i] );
    }
}

int main()
{
    /* We need to query for support of DXT1 textures. */
    char *extensions[1] = { "GL_NV_texture_compression_dxt1" };

	if ( !CreateGLWindow( 320, 240 ) )
	{
		FatalError( "Display initialization failed.\nExiting." );
	}

    /* Now that the context is created, verify the extension is supported. */
    VerifyExtensionsSupported( extensions, 1 );
    InitGraphicsState();

#ifndef _WIN32
    initRawKeyboard();
#endif

    /* Print Usage */
    fprintf( stderr, "\n" );
    fprintf( stderr, "Multitexture Keyboard Controls:\n" );
    fprintf( stderr, "------------------------------\n" );
    fprintf( stderr, "    keypad 1 -- No texture\n" );
    fprintf( stderr, "    keypad 2 -- Single texture: GL_REPLACE\n" );
    fprintf( stderr, "    keypad 3 -- Dual texture: GL_REPLACE, GL_REPLACE\n" );
    fprintf( stderr, "    keypad 4 -- Dual texture: GL_REPLACE, GL_MODULATE\n" );
    fprintf( stderr, "    keypad 5 -- Dual texture: GL_REPLACE, GL_BLEND\n" );
    fprintf( stderr, "    keypad 6 -- Dual texture: GL_REPLACE, GL_ADD\n" );
    fprintf( stderr, "    keypad 7 -- Dual texture: GL_MODULATE, GL_ADD\n" );
    fprintf( stderr, "    keypad 8 -- Dual texture: GL_BLEND, GL_ADD\n" );
    fprintf( stderr, "    keypad 9 -- Cycle to next texture environment combine mode\n" );
    fprintf( stderr, "    keypad 0 -- Toggle translation of texture 0\n" );
    fprintf( stderr, "    keypad . -- Toggle rotation of texture 1\n" );
    fprintf( stderr, "    <ESC>    -- Quit\n" );

    while ( !finished )
    {
#ifndef _WIN32
        GetKeyboardInput();
#endif
        keyboard( keypress );
        keypress = 255;
        tick();
    }

#ifndef _WIN32
	    closeRawKeyboard();
#endif

    CleanUp();

	return 0;
}

