/*
 * Simple app that draws a triangle using fixed-point entry points
 * into the driver and demonstrates how to allow for user interaction
 * using the keyboard.
 */

#include <GLES/egl.h>
#include <GLES/gl.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#ifndef _WIN32
#include "genericinput/genericinput.h"
#endif

static EGLDisplay       dpy;  /* handle to the display          */
static NativeWindowType win;  /* handle to the window           */
static EGLSurface       draw; /* handle to the drawable surface */
static EGLConfig        cfg;  /* handle to an available config  */
static EGLContext       ctx;  /* rendering context              */

#ifdef _WIN32
extern NativeWindowType MyWin32MakeWindow(NativeDisplayType disp,
                                          int x, int y, int w, int h);
#endif

/*
 * CreateGLWindow - open something to draw on
 */
int CreateGLWindow( int width, int height )
{
    EGLint num_cfgs = 0;
    int attrib[] =
    {
        EGL_RED_SIZE,   1, /* At least 1-bit red.   */
        EGL_GREEN_SIZE, 1, /* At least 1-bit green. */
        EGL_BLUE_SIZE,  1, /* At least 1-bit blue.  */
        EGL_DEPTH_SIZE, 1, /* At least 1-bit depth. */
        EGL_NONE
    };

    /* Get a handle to the default display. */
    dpy = eglGetDisplay( EGL_DEFAULT_DISPLAY );
    if (!dpy)
    {
        fprintf( stderr, "Error: Couldn\'t open the EGL default display\n" );
        return( EGL_FALSE );
    }

    /* Initialize the display before calling all other EGL functions. */
    if ( !eglInitialize( dpy, 0, 0 ) )
    {
        fprintf( stderr, "Error: Couldn\'t init the EGL default display\n" );
        return( EGL_FALSE );
    }

    /* Request an EGL config (i.e. pixel format) that has the requested attributes. */
    if ( !eglChooseConfig( dpy, attrib, &cfg, 1, &num_cfgs ) || (num_cfgs < 1) )
    {
        fprintf( stderr, "Error: Couldn\'t find a config on the EGL default display\n" );
        return( EGL_FALSE );
    }

#ifdef _WIN32
    /* To handle input on Windows, you must create your own window */
    /* with message processing and hand this window off to EGL.    */
    win = MyWin32MakeWindow( EGL_DEFAULT_DISPLAY, 351, 150, width, height );
#else
    /* To create a full screen window, query display width and height */
    /* and use them as parameters when making the window.             */
    width = eglDisplayWidthNV( dpy, cfg );
    height = eglDisplayHeightNV( dpy, cfg );
    win = eglMakeWindowNV( EGL_DEFAULT_DISPLAY, 0, 0, width, height );
#endif
    if (!win)
	{
        fprintf( stderr, "Error: Couldn\'t create window\n" );
        return( EGL_FALSE );
    }

    /* Create the drawable surface and rendering context. */
    draw = eglCreateWindowSurface( dpy, cfg, win, 0 );
    ctx = eglCreateContext( dpy, cfg, 0, 0 );

    /* Make the context current and bind the newly created drawable */
    /* surface as the current "draw" and "read" surface.            */
    eglMakeCurrent( dpy, draw, draw, ctx );

    return( EGL_TRUE );
}

/*
 * KillGLWindow - shut down rendering
 */
void KillGLWindow()
{
    if (dpy)
    {
        /* Delete created resources. */
        if (ctx) eglDestroyContext( dpy, ctx );
        if (draw) eglDestroySurface( dpy, draw );
        if (win) eglDestroyWindowNV( win );

        /* Release resources associated with the display. */
        eglTerminate( dpy );
    }
}

static void CleanUp()
{
	KillGLWindow();
}

#ifndef _WIN32
static void FatalError(const char *msg, ...)
     __attribute__ ((format (printf, 1, 2), noreturn));
#else
DECLSPEC_NORETURN static void FatalError(const char *msg, ...);
#endif

static void FatalError(const char *msg, ...)
{
    va_list ap;
#ifdef _WIN32
    char error[512];
#endif
    va_start(ap, msg);
    vfprintf(stderr, msg, ap);
    va_end(ap);
    fputc('\n', stderr);
#ifdef _WIN32
    va_start(ap, msg);
    _vsnprintf( error, sizeof(error), msg, ap ); 
    va_end(ap);
    MessageBox( NULL, error, "FatalError Message", MB_OK );
#endif
    CleanUp();
    exit(-1);
}

static void CheckError(const char *msg)
{
    int e = glGetError();
    if (e) FatalError("OpenGL ES Error %s: 0x%.8x", msg, e);
}

/* OpenGL ES supports 16.16 fixed-point interfaces.
 *
 * 16.16 fixed-point representation is defined in the OpenGL ES spec as
 * a 32-bit two's complement number with 16 bits of fraction.  Based on
 * this definition, floating-point and integer values may be converted
 * to fixed-point by taking the value multiplying it by 2^16 then converting
 * the resulting value to an integer.
 *
 * Using this approach:
 *     1.0 is 0x00010000 in fixed point
 *    -1.0 is 0xFFFF0000 in fixed point
 *    -0.5 is 0x00008000 in fixed point
 *    -0.5 is 0xFFFF8000 in fixed point
 *
 * When using fixed point representations of numbers, the rules for
 * multiplication and division are slightly different.  For this reason
 * when two fixed-point numbers are stored in variables of type 'int' it is
 * generally inappropriate to use the '/' and '*' operators when operating
 * on them.
 * 
 * Note: In the code below, "ROOT_3_OVER_2/3" produces the intended result
 * because the denominator is actually the integer value 3 -- not the 16.16
 * fixed- point representation of the value 3.
 *
 * For a general purpose math library, consider using the nvfxmath library
 * included with the NVIDIA Handheld SDK.  This library defines several
 * math functions for operating on 16.16 fixed-point types.
 */

#define FIXINT(x) (x << 16)
#define FIXHALF (FIXINT(1) >> 1)

#define ROOT_3_OVER_2 ((GLfixed) (unsigned) (0.8660254 * (1 << 16)))
#define ROOT_3_OVER_6 (ROOT_3_OVER_2/3)

static GLubyte ind [3] = { 0, 1, 2 };
static GLfixed vert[6] = { FIXHALF, -ROOT_3_OVER_6,
                          -FIXHALF, -ROOT_3_OVER_6,
                           FIXINT(0),  ROOT_3_OVER_2 - ROOT_3_OVER_6 };
static GLfixed tex [6] = { FIXINT(1), FIXINT(0),
                           FIXINT(0), FIXINT(0),
                           FIXHALF, FIXINT(1) };
static GLfixed col [12] = { FIXINT(1), FIXINT(0), FIXINT(0), FIXINT(1),
                            FIXINT(0), FIXINT(1), FIXINT(0), FIXINT(1),
                            FIXINT(0), FIXINT(0), FIXINT(1), FIXINT(1) };
static GLfixed angle = 0;
static int animate = 1;
static int finished = 0;

/*
 * On Windows, when a keypress occurs, the thread that is processing
 * messages will set this shared variable indicating which key was
 * most recently pressed.  This variable should use a semaphore for
 * exclusive access but since the message processing thread never
 * reads from this variable this is not critical.  In rare cases,
 * some keypress events will not be recognized by the main
 * rendering thread.
 *
 * There are alternative ways to handle keyboard input using this
 * EGL on windows implementation.  Please see the win32ui.c file in
 * this project to see the details of what the message processing
 * thread does.
 */
unsigned char keypress = 255;

void keyboard(unsigned char key)
{
    switch( key )
    {
        case ' ':
            animate = !animate;
            break;
        case 27:
            finished = 1;
            break;
    }
}

#ifndef _WIN32
static void GetKeyboardInput()
{
	unsigned char c;
	if ( readRawKeyboard(&c) > 0)
	{
	    keyboard( c );
	}
}
#endif

static void InitGraphicsState()
{
    glClearColorx(0, 0, 0, 0);
    glOrthox(-FIXINT(1), FIXINT(1), -FIXINT(1),
             FIXINT(1), -FIXINT(1), FIXINT(1));
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    glVertexPointer  (2, GL_FIXED, 0, (void *) vert);
    glTexCoordPointer(2, GL_FIXED, 0, (void *) tex);
    glColorPointer   (4, GL_FIXED, 0, (void *) col);

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	CheckError( "InitGraphicsState" );
}

static void Display(void)
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glLoadIdentity();
    glRotatex(angle, 0, 0, FIXINT(1));
    glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_BYTE, ind);
    eglSwapBuffers( dpy, draw );
}

static void tick(void)
{
    if ( animate )
        angle += FIXINT(15) / 10;

    Display();
}

int main()
{
	if ( !CreateGLWindow( 240, 320 ) )
	{
		FatalError( "Display initialization failed.\nExiting." );
	}

    InitGraphicsState();

#ifndef _WIN32
    initRawKeyboard();
#endif

    /* Print Usage */
    fprintf( stderr, "\n" );
    fprintf( stderr, "Simpleui Keyboard Controls:\n" );
    fprintf( stderr, "--------------------------\n" );
    fprintf( stderr, "    spacebar -- Toggle animation on/off\n" );
    fprintf( stderr, "    <ESC>    -- Quit\n" );

    while ( !finished )
    {
#ifndef _WIN32
        GetKeyboardInput();
#endif
        keyboard( keypress );
        keypress = 255;
        tick();
    }

#ifndef _WIN32
	    closeRawKeyboard();
#endif

    CleanUp();

	return 0;
}

