
#include "nvdds/nvdds.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// Defines
#define DDS_FOURCC            0x00000004l
#define DDS_PITCH             0x00000008l
#define DDS_COMPLEX           0x00000008l
#define DDS_TEXTURE           0x00001000l
#define DDS_MIPMAPCOUNT       0x00020000l
#define DDS_LINEARSIZE        0x00080000l
#define DDS_VOLUME            0x00200000l
#define DDS_MIPMAP            0x00400000l
#define DDS_DEPTH             0x00800000l

#define DDS_CUBEMAP           0x00000200L
#define DDS_CUBEMAP_POSITIVEX 0x00000400L
#define DDS_CUBEMAP_NEGATIVEX 0x00000800L
#define DDS_CUBEMAP_POSITIVEY 0x00001000L
#define DDS_CUBEMAP_NEGATIVEY 0x00002000L
#define DDS_CUBEMAP_POSITIVEZ 0x00004000L
#define DDS_CUBEMAP_NEGATIVEZ 0x00008000L

#define FOURCC_DXT1 0x31545844 //(MAKEFOURCC('D','X','T','1'))
#define FOURCC_DXT3 0x33545844 //(MAKEFOURCC('D','X','T','3'))
#define FOURCC_DXT5 0x35545844 //(MAKEFOURCC('D','X','T','5'))

typedef struct {
    unsigned short col0;
    unsigned short col1;
    unsigned char row[4];
} DXTColBlock;

typedef struct {
    unsigned short row[4];
} DXT3AlphaBlock;

typedef struct {
    unsigned char alpha0;
    unsigned char alpha1;
    unsigned char row[6];
} DXT5AlphaBlock;

typedef struct _DDS_PIXELFORMAT {
    unsigned long dwSize;
    unsigned long dwFlags;
    unsigned long dwFourCC;
    unsigned long dwRGBBitCount;
    unsigned long dwRBitMask;
    unsigned long dwGBitMask;
    unsigned long dwBBitMask;
    unsigned long dwABitMask;
} DDS_PIXELFORMAT;

typedef struct _DDS_HEADER {
    unsigned long dwSize;
    unsigned long dwFlags;
    unsigned long dwHeight;
    unsigned long dwWidth;
    unsigned long dwPitchOrLinearSize;
    unsigned long dwDepth;
    unsigned long dwMipMapCount;
    unsigned long dwReserved1[11];
    DDS_PIXELFORMAT ddspf;
    unsigned long dwCaps1;
    unsigned long dwCaps2;
    unsigned long dwReserved2[3];
} DDS_HEADER;

#define SWAP_CHAR(A, B)                         \
        do {                                    \
            unsigned char tmp;                  \
            tmp = A;                            \
            A = B;                              \
            B = tmp;                            \
        } while (0)

#define SWAP_SHORT(A, B)                        \
        do {                                    \
            unsigned short tmp;                 \
            tmp = A;                            \
            A = B;                              \
            B = tmp;                            \
        } while (0)

#ifdef MACOS
static void swap_endian(void *val)
{
    unsigned int *ival = (unsigned int *)val;

    *ival = ((*ival >> 24) & 0x000000ff) |
            ((*ival >>  8) & 0x0000ff00) |
            ((*ival <<  8) & 0x00ff0000) |
            ((*ival << 24) & 0xff000000);
}
#endif


static int size_dxtc(int width, int height, GLenum format)
{
    return ((width+3)/4)*((height+3)/4)*
        /* (format == GL_COMPRESSED_RGBA_S3TC_DXT1_EXT ? 8 : 16);   */
        (format == 0x83F0 ? 8 : 16);   
}


static void flip_blocks_dxtc1(DXTColBlock *line, int numBlocks)
{
    DXTColBlock *curblock = line;
    int i;

    for (i = 0; i < numBlocks; i++) {
        SWAP_CHAR(curblock->row[0], curblock->row[3]);
        SWAP_CHAR(curblock->row[1], curblock->row[2]);
        curblock++;
    }
}


#ifdef SUPPORT_DXT3
static void flip_blocks_dxtc3(DXTColBlock *line, int numBlocks)
{
    DXTColBlock *curblock = line;
    DXT3AlphaBlock *alphablock;
    int i;

    for (i = 0; i < numBlocks; i++) {
        alphablock = (DXT3AlphaBlock*)curblock;

        SWAP_SHORT(alphablock->row[0], alphablock->row[3]);
        SWAP_SHORT(alphablock->row[1], alphablock->row[2]);
        curblock++;

        SWAP_CHAR(curblock->row[0], curblock->row[3]);
        SWAP_CHAR(curblock->row[1], curblock->row[2]);
        curblock++;
    }
}
#endif


#if SUPPORT_DXT5
static void flip_dxt5_alpha(DXT5AlphaBlock *block)
{
    unsigned char gBits[4][4];
    
    const unsigned long mask = 0x00000007;          // bits = 00 00 01 11
    unsigned long bits = 0;
    memcpy(&bits, &block->row[0], sizeof(unsigned char) * 3);

    gBits[0][0] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[0][1] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[0][2] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[0][3] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[1][0] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[1][1] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[1][2] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[1][3] = (unsigned char)(bits & mask);

    bits = 0;
    memcpy(&bits, &block->row[3], sizeof(unsigned char) * 3);

    gBits[2][0] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[2][1] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[2][2] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[2][3] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[3][0] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[3][1] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[3][2] = (unsigned char)(bits & mask);
    bits >>= 3;
    gBits[3][3] = (unsigned char)(bits & mask);

    bits = (gBits[3][0] << 0) | (gBits[3][1] << 3) | (gBits[3][2] << 6) |
        (gBits[3][3] << 9) | (gBits[2][0] << 12) | (gBits[2][1] << 15) |
        (gBits[2][2] << 18) | (gBits[2][3] << 21);
    memcpy(&block->row[0], &bits, 3);

    bits = (gBits[1][0] << 0) | (gBits[1][1] << 3) | (gBits[1][2] << 6) |
        (gBits[1][3] << 9) | (gBits[0][0] << 12) | (gBits[0][1] << 15) |
        (gBits[0][2] << 18) | (gBits[0][3] << 21);
    memcpy(&block->row[3], &bits, 3);
}


static void flip_blocks_dxtc5(DXTColBlock *line, int numBlocks)
{
    DXTColBlock *curblock = line;
    DXT5AlphaBlock *alphablock;
    int i;

    for (i = 0; i < numBlocks; i++) {
        alphablock = (DXT5AlphaBlock*)curblock;
        
        flip_dxt5_alpha(alphablock);
        curblock++;

        SWAP_CHAR(curblock->row[0], curblock->row[3]);
        SWAP_CHAR(curblock->row[1], curblock->row[2]);
        curblock++;
    }
}
#endif


static void flip(char *image, int width, int height, int depth, int size,
                 GLenum format)
{
    int linesize, j;
    DXTColBlock *top;
    DXTColBlock *bottom;
    char *tmp;
    void (*flipblocks)(DXTColBlock*, int);
    int xblocks = width / 4;
    int yblocks = height / 4;
    int blocksize;

    switch (format) {
    /* case GL_COMPRESSED_RGBA_S3TC_DXT1_EXT: */
    case 0x83F0: 
        blocksize = 8;
        flipblocks = &flip_blocks_dxtc1; 
        break;
    /*
    case GL_COMPRESSED_RGBA_S3TC_DXT3_EXT: 
        blocksize = 16;
        flipblocks = &flip_blocks_dxtc3; 
        break;
    case GL_COMPRESSED_RGBA_S3TC_DXT5_EXT: 
        blocksize = 16;
        flipblocks = &flip_blocks_dxtc5; 
        break;
    */
    default:
        return;
    }

    linesize = xblocks * blocksize;
    tmp = malloc(linesize);

    for (j = 0; j < (yblocks >> 1); j++) {
        top    = (DXTColBlock*) (void*) (image + j * linesize);
        bottom = (DXTColBlock*) (void*) (image + (((yblocks-j)-1) * linesize));

        (*flipblocks)(top, xblocks);
        (*flipblocks)(bottom, xblocks);

        memcpy(tmp,    bottom, linesize);
        memcpy(bottom, top,    linesize);
        memcpy(top,    tmp,    linesize);
    }

    free(tmp);
}    


NVDDSImage *NVDDSLoad(char *filename)
{
    DDS_HEADER ddsh;
    char filecode[4];
    FILE *fp;
    NVDDSImage *image = NULL;
    int i, n;

    fp = fopen(filename, "rb");
    if (fp == NULL) {
        fprintf(stderr, "Could not open file %s.\n", filename);
        return NULL;
    }

    // read in file marker, make sure its a DDS file
    fread(filecode, 1, 4, fp);
    if (strncmp(filecode, "DDS ", 4) != 0) {
        fprintf(stderr, "%s not a DDS file.\n", filename);
        fclose(fp);
        return NULL;
    }

    image = calloc(sizeof(NVDDSImage), 1);

    // read in DDS header
    fread(&ddsh, sizeof(ddsh), 1, fp);

#ifdef MACOS
    swap_endian(&ddsh.dwSize);
    swap_endian(&ddsh.dwFlags);
    swap_endian(&ddsh.dwHeight);
    swap_endian(&ddsh.dwWidth);
    swap_endian(&ddsh.dwPitchOrLinearSize);
    swap_endian(&ddsh.dwMipMapCount);
    swap_endian(&ddsh.ddspf.dwSize);
    swap_endian(&ddsh.ddspf.dwFlags);
    swap_endian(&ddsh.ddspf.dwFourCC);
    swap_endian(&ddsh.ddspf.dwRGBBitCount);
    swap_endian(&ddsh.dwCaps1);
    swap_endian(&ddsh.dwCaps2);
#endif

    // check if image is a cubempa
    if (ddsh.dwCaps2 & DDS_CUBEMAP)
        image->cubemap = 1;

    // check if image is a volume texture
    if ((ddsh.dwCaps2 & DDS_VOLUME) && (ddsh.dwDepth > 0))
        image->volume = 1;

    // figure out what the image format is
    if (ddsh.ddspf.dwFlags & DDS_FOURCC) {
        switch(ddsh.ddspf.dwFourCC) {
        case FOURCC_DXT1:
            /* image->format     = GL_COMPRESSED_RGBA_S3TC_DXT1_EXT; */
            image->format     = 0x83F0;
            image->components = 3;
            image->compressed = 1;
            break;
        /*
        case FOURCC_DXT3:
            image->format     = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
            image->components = 4;
            image->compressed = 1;
            break;
        case FOURCC_DXT5:
            image->format     = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
            image->components = 4;
            image->compressed = 1;
            break;
        */
        default:
            fprintf(stderr, "Unsupported FOURCC code in %s.\n", filename);
            fclose(fp);
            return NULL;
        }
    } else {
        fprintf(stderr, "%s is not a DXTC image.\n", filename);
        fclose(fp);
        return NULL;
    }
    
    // store primary surface width/height/depth/numMipmaps
    image->width      = ddsh.dwWidth;
    image->height     = ddsh.dwHeight;
    image->depth      = ddsh.dwFlags&DDS_DEPTH       ? ddsh.dwDepth : 1;
    image->numMipmaps = ddsh.dwFlags&DDS_MIPMAPCOUNT ? ddsh.dwMipMapCount : 1;

    if (image->numMipmaps > NVDDS_MAX_MIPMAPS) {
        fprintf(stderr,
                "Too many mipmaps, increase NVDDS_MAX_MIPMAPS in nvdds.h.\n");
        fclose(fp);
        return NULL;
    }
    
    // load all surfaces for the image (6 surfaces for cubemaps)
    for (n=0; n < (image->cubemap ? 6 : 1); n++) {
        int width  = image->width;
        int height = image->height;

        for (i = 0; i < image->numMipmaps; i++) {
            // Get the size, read in the data.
            int size = size_dxtc(width, height, image->format);
            image->data[i] = malloc(size);
            image->size[i] = size;
            fread(image->data[i], size, 1, fp);

            // Flip in Y for OpenGL
            flip(image->data[i], width, height, image->depth, size,
                 image->format);

            // shrink to next power of 2
            width  >>= 1;
            if (width < 4) {
                width = 4;
            }
            height >>= 1;
            if (height < 4) {
                height = 4;
            }
        }
    }

    // swap cubemaps on y axis (since image is flipped in OpenGL)
    if (image->cubemap) {
        int y1 = 2*image->numMipmaps;
        int y2 = 3*image->numMipmaps;
        for (i=0; i<image->numMipmaps; i++, y1++, y2++) {
            void *tmp = image->data[y1];
            image->data[y1] = image->data[y2];
            image->data[y2] = tmp;
        }
    }
    
    fclose(fp);

    return image;
}


void NVDDSFree(NVDDSImage *image)
{
    int i;

    for (i=0; i<image->numMipmaps; i++) {
        free(image->data[i]);
    }
    free(image);
}

